/*
Auteur          : Pierre Larbier
Developp le    : 7.12.1993
Derniere MAJ    : 21.11.1994
Toutes les routines de ce fichier sont librement utilisables
PF rajoute +1  jour2sem: 9 mai 2002
*/
#include <conio.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <dos.h>

#define OK      0
#define ERREUR  1

const char *nom_jour[] = {
	"Lundi",
	"Mardi",
	"Mercredi",
	"Jeudi",
	"Vendredi",
	"Samedi",
	"Dimanche"
	};

const char *nom_mois[] = {
	"janvier",
	"fevrier",
	"mars",
	"avril",
	"mai",
	"juin",
	"juillet",
	"aout",
	"septembre",
	"octobre",
	"novembre",
	"decembre"
	};

const char *rom_chiffre[]  = { "I","V","X","L","C","D","M","|))","((|))"};
const char deb_rep[] = {22,22,22,23,22,22,22,23,23,23,23,24,23,23,23};

/* Calcule le jour julien d'une date  */
long jour2jul(char jour , char mois , short annee);
/* Verifie l'existence d'une date specifiee en entree */
char existj(char jour , char mois , short annee);
/* Calcule la date d'un jour julien */
char jul2jour(char *jour , char *mois , short *annee , long jule);
/* Calcule date1-date2 en Nb de jours */
long ecart_date(char j1 , char m1 , short a1 , char j2 , char m2 , short a2);
/* Calcule la nouvelle date de date=date + Nb jours */
char ajoute_jour(char *j, char *m, short *a, long n);
/* Calcule le nom et le numero du jour de la semaine d'un jour julien donn */
char *j_txt(long n , char *num);
/*  Renvoie le n de la semaine dans l'anne d'une date spcifie */
char jour2semaine(char jour , char mois , short annee);
/* Renvoie la date du lundi d'une semaine spcifie en entre */
char semaine2jour(char *jour , char *mois , short *annee , char semaine);
/* Place une date dans une chaine de caractere fournie */
char date2asc(char jour , char mois , short annee , char *txt);
/* Convertit une date gregorienne en republicaine */
char gregor2rep(char *jour , char *mois , short *annee);
/* Convertit une date republicaine en gregorienne */
char rep2gregor(char *jour , char *mois , short *annee);
/* Convertit un nombre en chiffre romains */
char chiffre2rom(short chiffre , char *romain);

/*******************************************************************
  Procdure : jour2jul
  Auteur / Date : Pierre Larbier 7/11/1994
  Fonction : Calcule le n du jour julien d'une date donne.
     Le jour n0 correspond au Lundi 1er Janvier 4713 av. JC.
     Nous sommes aujourd'hui le 7/12/1993 au jour n 2 449 329.
     Cette fonction prend en compte :
	- le fait que l'anne 0 n'existe pas
	- le changement de calendrier le 4 octobre 1582 (le lendemain
	  du Jeudi 4 octobre 1582 fut le vendredi 15 octobre 1582)
	- les annes bissextiles sont conformes au calendrier julien
	  avant le 4 octobre 1582 et conformes au calendrier grgorien
	  aprs le 15 octobre 1582.
     Cette fonction est suceptible de renvoyer n'importe quoi si la
     date spcifie n'existe pas. Cependant, elle a la bonne ide
     de considrer que le mois n13 est le mois de janvier de l'anne
     suivante (ce qui est utilis dans la routine existj().
     Les calculs sont effectus entirement en entier.
  Entres : jour, mois , annee : date dont on cherche le jour julien.
  Sortie  : jour julien correspondant  la date.

*************************************************************************/
long jour2jul(char jour , char mois , short annee)
{
long jule, gregorien , a , j , m;
/* Attention, cette routine ne doit etre appellee que si on est sur
   que le jour fourni en parametre existe !!!!!! */
a = (long)annee;
m = (long)mois;
j = (long)jour;
/* On commence par faire un ajustement pour les annees negatives */
	if (annee < 0)	a++;
	/* Cette 1ere formule tient compte des annees bissextiles "normales"
	   c'est  dire des annees divisibles par 4 (calendrier Julien) */
	jule = ((489*m-481)>>4) + j + 1721423L - (((a&3)+7)>>2)*((m+13)>>4)
				+ ((1461*(a-1))>>2);
	/* Avant le 04/10/1582 on est en calendrier julien, et en calendrier
	   gregorien aprs le 15/10/1582 */
	if (jule <= 2299160L)
		return jule;
	/* En calendrier gregorien les ajustements  faire sont :
	   - les annees divisibles par 100 ne sont pas bissextiles
	   - les annees divisibles par 400 sont bissextiles.
	   J'ai laisse l'ajustement en clair pour que vous puissiez integrer
	   facilement la modification des 4000 ans */
	 gregorien = jule - (a/100L - a/400L - 2L);
	 if ( ( (a%100) == 0 ) && (m < 3) && ( (a%400) != 0) )
		gregorien++;
	return gregorien;
}

/*************************************************************************

  Procdure : existj
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Teste si une date specifiee est valide (par exemple le 53/21/1993
	     n'existe pas)
  Entres : jour, mois , annee : date dont on vrifie l'existence.
  Sortie  : OK si la date existe
	    ERREUR si :
		- le jour n'est pas dans l'intervalle [1,31]
		- le mois n'est pas dans l'intervalle [1,12]
		- l'annee est nulle
		- la date est entree le 5/10/1582 et le 14/10/1582 inclus
		- le jour est plus grand que le nombre de jour du mois en
		  question.

*************************************************************************/
char existj(char jour , char mois , short annee)
{
short nbj;
	/* On commence par liminer le plus gros */
	if ((jour < 1) || (jour > 31) || (mois < 1) || (mois > 12) || (annee == 0))
		return ERREUR;
	/* Puis le cas (tres rare normalement) des 10 jours  du 4 au 15 octobre 1582
	qui ont disparu quand on est passe du calendrier Julien au Gregorien */
	if ( ( annee == 1582) && (mois == 10) ) {
		if ( (jour > 4) && (jour < 15) )
			return ERREUR;
		else
			return OK;
	  }
	/* On regarde ensuite combien il y a de jour dans le mois en question grace
	    la fonction jour2jul(). A noter que jour2jul() marche pour notre
	   application si on lui passe 13 comme num de mois) */

	nbj = (short)jour2jul( 1 , mois + 1 , annee) - (short)jour2jul(1,mois,annee);
	/* Et si le jour propose est superieur au nbre de jours de ce mois l,
	   on dit qu'il n'existe pas */
	if (jour > nbj)
		return ERREUR;

	return OK;
}

/*************************************************************************

  Procdure : jul2jour
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Renvoie la date correspondant  un jour julien specife en entree.
	     C'est la fonction inverse de jour2jul().
	     L'algorithme utilise ici n'est pas tres brillant (c'est le moins
	     qu'on puisse dire) mais on ne travaille qu'en entier et pour
	     l'usage qu'on peut en faire, sa lenteur ne doit pas etre un
	     handicap trop grave.
	     J'ai verifie que dans la gamme de dates possibles (du 1/1/-32768
	     au 31/12/32767) cette fonction concorde bien avec jour2jul().
  Entrees : jour, mois , annee : pointeurs sur les elements de la date recherchee
	    jule : jour julien dont on veut la date correspondante.
  Sortie  : OK si la date produite est valide.
	    ERREUR si la date produite sort de l'intervalle
			 [1/1/-32768,31/12/32767]

*************************************************************************/
char jul2jour(char *jour , char *mois , short *annee , long jule)
{
long an , deb , fin;
char m , j;
/* Si l'annee que l'on va trouver n'est pas representable dans un  short on sort en erreur */
	if ((jule < -10247089L) || (jule > 13689325L))	return ERREUR;
	/* On commence par chercher une premire estimation de l'annee */
	/* Cas de cette fichue annee 1582 */
	if ((jule >= 2298884L) && (jule <= 2299238L)) {
		an = 1582L;
		deb = 2298884L;
	}
	else {
		if (jule > 2299160L) {
			/* Cas des jours gregoriens */
			an = (4L*(jule - 2299239L))/1461L;
			/* Ajustement tenant compte des annees de siecle
			   non bissextiles (inutile quand on se limite
			    + ou - 32767 annees) */
			an = an + (3L*an)/146100L + 1583L;
		}
		else
			/* Cas des jours juliens */
			an = (4L*jule)/1461L - 4713L;
		if ((an >= 0L) && (an < 1583L))	an++;
		deb = jour2jul(1,1,(short)an);
		fin = jour2jul(1,1,(short)an+1);
		if (deb == fin)
			fin = jour2jul(1,1,(short)an+2);
		else {
			if (jule >= fin) {
				an++;
				deb = fin;
			}
			else {
				if (jule < deb) {
					an--;
					deb = jour2jul(1,1,(short)an);
				}
			}
		}
	}
	*annee = (short)an;
	/* On dispose a present de l'annee et de son jour de debut, on va
	   s'en servir pour trouver le mois et le jour */
	m = (short)((jule - deb)/31L) + 1;
	deb = jour2jul(1 , m+1 , (short)an);
	if (jule >= deb) m++; else deb = jour2jul(1 ,m , (short)an);
	j = (short)(jule - deb + 1L);
	/* Il ne nous reste plus que le cas vicieux d'octobre 1582 */
	if ((an == 1582L) && (m == 10) && (j > 4)) j = j + 10;
	*mois = m;
	*jour = j;
	return OK;
}

/*************************************************************************

  Procdure : ecart_date
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Calcule le nombre de jours entre deux dates.
	     C'est une utilisation directe du jour julien.
	     Les deux dates doivent tre valides, sans quoi la fonction risque
	     de renvoyer n'importe quoi.
  Entres : j1 , m1 , a1 : date de dpart
	    j2 , m2 , a2 : date d'arrive
  Sortie  : nombre de jours entre les deux dates. Si j1/m1/a1 est le 7/12/1993
	    et j2/m2/a2 est le 8/12/1993 la fonction renvoie -1L.

*************************************************************************/
long ecart_date(char j1 , char m1 , short a1 , char j2 , char m2 , short a2)
{
	return jour2jul(j1 , m1 , a1) -  jour2jul(j2 , m2 , a2);
}

/*************************************************************************

  Procdure : ajoute_jour
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Calcule la date obtenue en ajoutant un nombre de jour specifie
	      une date donnee.
  Entres : jour , mois , annee : pointeurs sur les elements de la date de
				  depart et d'arrivee.
	    n : nombre de jours  ajouter (n peut tre ngatif ou nul).
  Sortie  : OK ou ERREUR si :
		- la date de depart n'est pas valide.
		- la date d'arrivee sort de l'intervalle
		  [1/1/-32768,31/12/32767].

*************************************************************************/
char ajoute_jour(char *j, char *m, short *a, long n)
{
	/* Teste la date de depart */
	if (existj(*j , *m , *a) != OK)
	 return ERREUR;
	/* Calcule la date finale obtenue */
	return  jul2jour(j, m, a, n + jour2jul(*j , *m , *a));;
}

/*************************************************************************

  Procdure : j_txt
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Calcule le numero du jour de la semaine d'un jour julien donne.
	     Les jours sont cods :
		0 : lundi
		1 : mardi
		2 : mercredi
		3 : jeudi
		4 : vendredi
		5 : samedi
		6 : dimanche
	     La fonction renvoie en plus un pointeur sur une chaine de
	     caractre contenant le nom du jour.
  Entres : n : jour julien dont on cherche le rang dans la semaine
	    num : pointeur sur un char o la routine place le n du jour de
		  la semaine calcul.
  Sortie  : pointeur sur une chaine de caractre contenant le nom du jour de
	    la semaine.

*************************************************************************/
char *j_txt(long n , char *num)
{
char c;
	if (n < 0L)
		c = 6 - (char)((-n-1)%7);
	else
		c = (char)(n%7);
	*num = c;
	return (char *)nom_jour[c];
}

/*************************************************************************

  Procdure : jour2semaine
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Renvoie le num de la semaine dans l'annee d'une date spcifiee.
	     Le numro de semaine est compris entre 1 et 53.
	     Dans le cas rarissime ou la date specifiee est avant le 1er lundi
	     de l'annee, on est en fait dans la dernire semaine de l'annee
	     precedente.
  Entres : jour , mois , annee : elements de la date dont on cherche le
		numero de la semaine.
  Sortie  : numero de la semaine calcul.

*************************************************************************/
char jour2semaine(char jour , char mois , short annee)
{
long debut, courant , n;
	/* Commence par regarder si le jour en question existe. Sinon, on
	sort en renvoyant 0 (ce qui n'est pas un num de semaine possible) */
	if (existj(jour , mois , annee) != OK)
		return 0;
	/* Puis le num du jour specifie */
	courant = jour2jul(jour , mois , annee);
	/* On recherche ensuite le jour julien correspondant au lundi de
	   la semaine n1 de l'annee specifiee */
	debut = jour2jul(1 , 1 , annee);
	if (debut < 0L)
		n = 6L - (-debut-1)%7;
	else
		n = debut%7;
	if (n != 0)
		debut += 7 - n;

	/* Au cas extremement improbable ou le jour specifie se trouverait
	   avant le premier lundi de l'annee, on est la semaine 52 ou 53*/
	if (courant < debut) {
		if (annee == 1)
			n = -1;
		else
			n = annee - 1;
		debut = jour2jul(1 , 1 , (short)n);
		if (debut < 0L)
			n = 6L - (-debut-1)%7;
		else
			n = debut%7;
		if (n != 0)
			debut += 7 - n;
	}
/* Il ne reste plus qu'a faire la division entiere (on beneficie
   de la bonne idee de Grgoire XIII de passer d'un jeudi au vendredi
   donc on a pas a tester le cas d'octobre 1582) */
  n=(courant - debut)/7 + 2;   /* modif 09 mai 2002 was +1 */
  if (n>53) n=1;  /* modif 09 mai 2002 traitement du 1/1/02 */
 return n; 
}

/*************************************************************************

  Procdure : semaine2jour
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Renvoie la date du lundi d'une semaine spcifie en entre.
	     Le numro de semaine est compris entre 1 et 53.
	     Cette fonction est quasiement l'inverse de jour2semaine().
  Entres : jour , mois , annee : pointeurs lments de la date correspondant
		 la semaine spcifie. A l'appel de la fonction seul *annee
		est utilis. La routine renseigne jour et mois.
	    semaine : numro de la semaine spcifi

  Sortie  : OK ou ERREUR si :
		- la semaine spcifie n'existe pas
		- l'anne spcifie n'existe pas

*************************************************************************/
char semaine2jour(char *jour , char *mois , short *annee , char semaine)
{
long debut, n , suivant;
short an;

	/* Premier test d'erreur */
	if ((semaine < 1) || (semaine > 53) || (*annee == 0))
		return ERREUR;

	/* On le jour julien correspondant au lundi de la semaine num 1 de l'annee specifiee */
	debut = jour2jul(1 , 1 , *annee);
	if (debut < 0L)
		n = 6L - (-debut-1)%7;
	else
		n = debut%7;
	if (n != 0)
		debut += 7 - n;

	/* Afin de faire la derniere verification, recherche le jour julien
	   correspondant au premier lundi de l'annee suivante */
	an = *annee + 1;
	if (an == 0)
		an = 1;
	suivant = jour2jul(1 , 1 , an);
	if (debut < 0L)
		n = 6L - (-suivant-1)%7;
	else
		n = suivant%7;
	if (n != 0)
		suivant += 7 - n;
	if (((suivant - debut)/7) < semaine)
		return ERREUR;

	/* Et on sort en renvoyant la date du lundi de la semaine specifiee */
	jul2jour(jour, mois , annee , debut);
	ajoute_jour(jour, mois, annee, 7*(semaine-1));

	return OK;
}

/*************************************************************************
  Procdure : date2asc
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Remplit une chaine de caracteres avec la date fournie en
		entree. Le format est le suivant :
		nom du jour (en lettres), jour (1 ou 2 chiffres) , nom du
		mois (en lettres) , annee (nombre positif , 5 chiffres au
		maximum) , "av. JC" si l'annee est negative.
		Par exemple, la routine produit la chaine suivante pour le
		7/12/1993 : "Mardi 7 decembre 1993".
		La plus longue chaine produite fait 34 caractres, donc il
		faut que le programme appelant reserve 35 octets pour la
		contenir (35  cause du \0 terminal).
  Entrees : jour , mois , annee : date a convertir en ascii
	    txt : pointeur sur la chaine de caracteres a renseigner.
  Sortie  : OK ou ERREUR si :
		- la date specifiee n'existe pas.
*************************************************************************/
char date2asc(char jour , char mois , short annee , char *txt)
{
char msg[35] , len;
char c;
	/* Si la date specifiee n'existe pas, on sort en erreur */
	if (existj(jour, mois, annee) != OK)
		return ERREUR;

	/* Constitue la chaine ascii representant la date */
	msg[0] = 0;
	strcat(msg , j_txt(jour2jul(jour , mois , annee) , &c));
	len = strlen(msg);
	if (annee < 0)
		sprintf(&msg[len]," %d %s %d av. JC",jour,nom_mois[mois-1],-annee);
	else
		sprintf(&msg[len]," %d %s %d",jour,nom_mois[mois-1],annee);

	/* recopie la chaine obtenue */
	strcpy(txt, msg);
	return OK;
}

/*************************************************************************
  Procdure : gregor2rep
  Auteur / Date : Pierre Larbier 21/11/1994
  Fonction : Convertit une date gregorienne en date republicaine.
		Elle fonctionne pour les dates comprises entre le 22/9/1792
		et le 31/12/1805 inclus.
		Les numeros des mois republicains sont les suivants :
		vendmiaire : 1
		brumaire    : 2
		frimaire    : 3
		nivse      : 4
		pluvise    : 5
		ventse     : 6
		germinal    : 7
		floral     : 8
		prairail    : 9
		messidor    : 10
		thermidor   : 11
		fructidor   : 12
		sans-culottides : 13 (ce n'est pas un mois  proprement
			     parler mais il faut bien les mettre
			     quelque-part).

  Entres : jour , mois , annee : pointeurs sur la date pour les convertir.
			en cas d'erreur ils ne sont pas modifis.
  Sortie  : OK ou ERREUR si :
		- la date specifiee n'existe pas dans le calendrier
		  republicain.

*************************************************************************/
char gregor2rep(char *jour , char *mois , short *annee)
{
long n;
short i;
	/* Commence par regarder si la date existe dans le calendrier
	   republicain */
	if (existj(*jour, *mois, *annee) != OK)
		return ERREUR;
	n = jour2jul(*jour, *mois, *annee);
	if ((n < 2375840L) || (n > 2380687L))
		return ERREUR;

	/* Cherche l'annee sans aucun effort d'optimisation... desole */
	i = 1;
	while ((jour2jul(deb_rep[i-1], 9, i+1791) <= n) && (i != 15))
		i++;
	*annee = i-1;

	/* On en deduit le mois et le jour */
	n = n - jour2jul(deb_rep[i-2], 9, i+1790);
	*mois = 1 + n/30;
	*jour = 1 + n - (*mois - 1)*30;
	/* Et c'est fini */
	return OK;
}
/*************************************************************************
  Procdure : rep2gregor
  Auteur / Date : Pierre Larbier 21/11/1994
  Fonction : Convertit une date republicaine en date gregorienne
		Elle fonctionne pour les dates comprises entre le
		1er vendemiaire an I et le 10 nivose an XIV inclus.
		Les numeros des mois rpublicains sont les suivants :
		vendemiaire : 1
		brumaire    : 2
		frimaire    : 3
		nivose      : 4
		pluviose    : 5
		ventose     : 6
		germinal    : 7
		floreal     : 8
		prairail    : 9
		messidor    : 10
		thermidor   : 11
		fructidor   : 12
		sans-culottides : 13 (ce n'est pas un mois  proprement
			     parler mais il faut bien les mettre
			     quelque-part).
  Entrees : jour , mois , annee : pointeurs sur la date pour les convertir.
			en cas d'erreur ils ne sont pas modifies.
  Sortie  : OK ou ERREUR si :
		- la date specifiee n'existe pas dans le calendrier  republicain.

*************************************************************************/
char rep2gregor(char *jour , char *mois , short *annee)
{
long n;
	/* Commence par regarder si la date existe */
	if ((*annee < 1) || (*annee > 14))
		return ERREUR;
	if (*jour > 30)
		return ERREUR;
	if (*annee == 14) {
		if (*mois > 4)
			return ERREUR;
		if ((*mois == 4) && (*jour > 10))
			return ERREUR;
	}
	if (*mois == 13) {
		if (*jour > 6)
			return ERREUR;
		if ((*jour == 6)&&(*annee!=3)&&(*annee!=7)&&(*annee!=11))
			return ERREUR;
	}
	/* Et on converti la date */
	n = jour2jul(deb_rep[*annee-1], 9, *annee+1791);
	n = n + (*mois-1)*30 + *jour - 1;
	jul2jour(jour , mois , annee , n);

	/* Et voil le travail */
	return OK;
}

/*************************************************************************
  Procdure : chiffre2rom
  Auteur / Date : Pierre Larbier 7/12/1993
  Fonction : Convertit un short en chiffres romains. Le nombre  convertir
		doit etre strictement positif.
		La syntaxe utilisee pour le chiffre romain est celle employee
		courament aujourd'hui :
		- un chiffre dcimal correspond  un ensemble de symboles
		  romains (4 symboles au maximum pour le 8)
		- l'emploi de la synthese soustractive est limit au 4 et au  9.
		- On utilise les symboles "|))" et "((|))" pour representer
		  respectivement 5000 et 10 000 (les symboles superieurs ne
		  sont pas necessaires puisqu'on est limit  32767).
		De ce fait, 84 sera code "LXXXIV" et non pas "XXCIIII".
		De meme, 1999 sera cod "MCMXCIX" et pas "MIM" par exemple.
		La plus longue chaine produite fait 28 caractres (pour le
		nombre 28 888), le programme appelant devra donc reserver 
		29	octets (29  cause du \0 terminal).
  Entrees : jour , mois , annee : date  convertir en ascii
	    txt : pointeur sur la chaine de caractres  renseigner.
  Sortie  : OK ou ERREUR si :
		- le nombre est inferieur  1.
*************************************************************************/
char chiffre2rom(short chiffre , char *romain)
{
char msg[29], dec[6] , i;

	/* Si le chiffre specifi est negatif ou nul, on sort en erreur */
	if (chiffre <= 0)
		return ERREUR;
	/* Dcompose le chiffre en base 10 */
	i = 0;
	do {
		dec[i++] = chiffre % 10;
		chiffre = chiffre / 10;
	} while (chiffre != 0);

	/* Construit le chiffre romain */
	msg[0] = 0;
	do {
		switch(dec[--i]) {
			case 0: break;
			case 3: strcat(msg , rom_chiffre[2*i]);
			case 2: strcat(msg , rom_chiffre[2*i]);
			case 1: strcat(msg , rom_chiffre[2*i]); break;
			case 5: strcat(msg , rom_chiffre[2*i+1]); break;
			case 6: strcat(msg , rom_chiffre[2*i+1]);
				strcat(msg , rom_chiffre[2*i]); break;
			case 7: strcat(msg , rom_chiffre[2*i+1]);
				strcat(msg , rom_chiffre[2*i]);
				strcat(msg , rom_chiffre[2*i]); break;
			case 8: strcat(msg , rom_chiffre[2*i+1]);
				strcat(msg , rom_chiffre[2*i]);
				strcat(msg , rom_chiffre[2*i]);
				strcat(msg , rom_chiffre[2*i]); break;
			case 9: strcat(msg , rom_chiffre[2*i]);
				strcat(msg , rom_chiffre[2*i+2]); break;
			case 4: strcat(msg , rom_chiffre[2*i]);
				strcat(msg , rom_chiffre[2*i+1]); break;
		}
	} while (i > 0);
	/* recopie la chaine obtenue */
	strcpy(romain, msg);
	return OK;
}
